@@ -35,6 +35,7 @@
 #include <com/sun/star/frame/XDispatchProvider.hpp>
 #include <com/sun/star/frame/XLayoutManager.hpp>
 #include <com/sun/star/frame/XModuleManager.hpp>
+#include <com/sun/star/frame/XPopupMenuController.hpp>
 #include <com/sun/star/frame/status/Visibility.hpp>
 #include <com/sun/star/lang/SystemDependent.hpp>
 #include <com/sun/star/beans/PropertyValue.hpp>
@@ -65,24 +66,27 @@ using com::sun::star::frame::XDispatchProvider;
 using com::sun::star::frame::XDispatchHelper;
 using com::sun::star::frame::XModuleManager;
 using com::sun::star::frame::XLayoutManager;
+using com::sun::star::frame::XPopupMenuController;
 using com::sun::star::frame::FeatureStateEvent;
 using com::sun::star::frame::status::Visibility;
 using com::sun::star::lang::SystemDependent::SYSTEM_XWINDOW;
 using com::sun::star::uno::UNO_QUERY;
 using com::sun::star::uno::UNO_QUERY_THROW;
 using com::sun::star::uno::Sequence;
+using com::sun::star::uno::XComponentContext;
 using com::sun::star::ui::XUIElement;
 using com::sun::star::util::URL;
 using com::sun::star::util::XURLTransformer;
 
-gboolean
-item_about_to_show (DbusmenuMenuitem *item, gpointer user_data);
+// Item callbacks
+void     item_activated     (DbusmenuMenuitem *item, guint timestamp, gpointer user_data);
+gboolean item_about_to_show (DbusmenuMenuitem *item, gpointer user_data);
 
-typedef enum {
-	MENU_ITEM_INVALID = 0,
-	MENU_ITEM_ENABLED = 1,
-	MENU_ITEM_DISABLED = 2
-} MenuItemStatus;
+void
+destroy_menuitem (gpointer data)
+{
+	g_object_unref (G_OBJECT (data));
+}
 
 // This is used in a hash table with commandurls as keys
 class MenuItemInfo {
@@ -206,7 +210,7 @@ class MenuItemStatusListener : public cppu::WeakImplHelper1 < XStatusListener >
 		gchar* c_url = g_utf16_to_utf8 (url.getStr(),
 			                            url.getLength(),
 			                            NULL, NULL, NULL);
-		g_debug ("UPDATE %s", c_url);
+//		g_debug ("UPDATE %s", c_url);
 		
 		GHashTable *commandsInfo = helper->getCommandsInfo ();
 		MenuItemInfo *info = (MenuItemInfo*)g_hash_table_lookup (commandsInfo, (gpointer)c_url);
@@ -265,16 +269,6 @@ class MenuItemStatusListener : public cppu::WeakImplHelper1 < XStatusListener >
 	{}
 };
 
-
-// Item callbacks
-void item_activated (DbusmenuMenuitem *item, guint timestamp, gpointer user_data);
-
-void
-destroy_menuitem (gpointer data)
-{
-	g_object_unref (G_OBJECT (data));
-}
-
 FrameHelper::FrameHelper(const Reference< XMultiServiceFactory >&  rServiceManager,
                     const Reference< XFrame >&        xFrame,
                     DbusmenuServer*                   server)
@@ -299,6 +293,25 @@ FrameHelper::FrameHelper(const Reference< XMultiServiceFactory >&  rServiceManag
 		                                  g_str_equal,
 		                                  g_free,
 		                                  destroy_menu_item_info);
+		                                  
+		                                  
+   	xMM = Reference < XModuleManager> (xMSF->createInstance(OUString::createFromAscii("com.sun.star.frame.ModuleManager")),
+	                                   UNO_QUERY);
+	xPCF = Reference < XMultiComponentFactory > (xMSF->createInstance(OUString::createFromAscii("com.sun.star.frame.PopupMenuControllerFactory")),
+	                                             UNO_QUERY);
+	
+	args = Sequence < Any > (2);
+	//TODO: The frame identifier should be updated onLoad
+	
+	PropertyValue item;
+	
+	item.Name = OUString::createFromAscii("ModuleName");
+	item.Value <<= xMM->identify (xFrame);
+	args[0] <<= item;
+	
+	item.Name = OUString::createFromAscii("Frame");
+	item.Value <<= xFrame;
+	args[1] <<= item;
 	
 	root = NULL;
 	watcher_set = FALSE;
@@ -437,6 +450,9 @@ FrameHelper::rebuildMenu (Reference < XMenu >  xMenu,
 		if (!DBUSMENU_IS_MENUITEM (item))
 			continue;
 		
+		// We drop the WindowList, it's useless and doesn't work properly
+		if (oUCommand.equals (OUString::createFromAscii (".uno:WindowList")))
+			continue;
 		
 		//We set the default properties (in case it was not visible or a separator)
 		dbusmenu_menuitem_property_set (item, DBUSMENU_MENUITEM_PROP_TYPE, DBUSMENU_CLIENT_TYPES_DEFAULT);
@@ -453,11 +469,8 @@ FrameHelper::rebuildMenu (Reference < XMenu >  xMenu,
 		
 		
 		// Replace the tilde with an underscore 
-	
 		OString command = OUStringToOString (oUCommand, RTL_TEXTENCODING_ASCII_US);
-		dbusmenu_menuitem_property_set (item, "CommandURL", command.getStr ());
-		
-		//g_debug ("new item: %s", command.getStr ());
+		dbusmenu_menuitem_property_set (item, "CommandURL", command.getStr());
 
 		MenuItemInfo* commInfo = (MenuItemInfo*)g_hash_table_lookup (commandsInfo, (gconstpointer)command.getStr());
 		if (!commInfo)
@@ -467,13 +480,18 @@ FrameHelper::rebuildMenu (Reference < XMenu >  xMenu,
 			
 			
 			OUString oULabel = getLabelFromCommandURL(oUCommand);
+			if (oULabel.getLength() == 0)
+			{
+				oULabel = xMenu->getItemText (id);
+			}
+			
 			//Replace tilde with underscore for Dbusmenu Alt accelerators
 			oULabel = oULabel.replace ((sal_Unicode)0x007e, (sal_Unicode)0x005f);
 	
 			// GLib behaves better than OUStringToOString wrt encoding transformation
 			gchar* label = g_utf16_to_utf8 (oULabel.getStr(),
 			                                oULabel.getLength(),
-			                                NULL, NULL, NULL);	
+			                                NULL, NULL, NULL);
 			commInfo->setLabel (label);
 					
 			g_free (label);
@@ -504,16 +522,44 @@ FrameHelper::rebuildMenu (Reference < XMenu >  xMenu,
 		Reference < XDispatch > xDispatch  = xdp->queryDispatch (commandURL, OUString(), 0);
 		if (xDispatch.is())
 			xDispatch->addStatusListener (xSL, commandURL);
-		
-		// Introspect submenus
+
 		Reference < XPopupMenu > subPopMenu (xMenu->getPopupMenu (id), UNO_QUERY);
-		if (subPopMenu.is ())
+		
+		//Build special submenu information
+		if (isSpecialSubmenu (oUCommand))
 		{
-			Reference <XMenu> subMenu (subPopMenu, UNO_QUERY);	
-			g_signal_connect(G_OBJECT(item), DBUSMENU_MENUITEM_SIGNAL_ABOUT_TO_SHOW, G_CALLBACK(item_about_to_show), this);			
+			Reference < XPropertySet > xMSFProps (xMSF, UNO_QUERY);
+			Reference <XComponentContext> xContext (xMSFProps->getPropertyValue (OUString::createFromAscii ("DefaultContext")),
+			                                        UNO_QUERY);
+
+			Reference < XPopupMenuController > xRFC (xPCF->createInstanceWithArgumentsAndContext(oUCommand,
+	                                                                                             args,
+                                                                                                 xContext),
+                                                     UNO_QUERY);
+
+			Reference < XPopupMenu > xPO (xMSF->createInstance(OUString::createFromAscii ("stardiv.Toolkit.VCLXPopupMenu")),
+			                              UNO_QUERY);
+			
+			if (xRFC.is () && xPO.is ())
+			{
+				xRFC->setPopupMenu (xPO);
+				xRFC->updatePopupMenu ();
+				Reference < XMenu > subMenu (xPO, UNO_QUERY);
+				rebuildMenu (subMenu, item);
+			}
+			else if (subPopMenu.is ())
+			{
+				Reference <XMenu> subMenu (subPopMenu, UNO_QUERY);
+				rebuildMenu (subMenu, item);			
+			}
+		}
+		// Introspect submenus
+		else if (subPopMenu.is ())
+		{
+			Reference <XMenu> subMenu (subPopMenu, UNO_QUERY);
+			g_signal_connect(G_OBJECT(item), DBUSMENU_MENUITEM_SIGNAL_ABOUT_TO_SHOW, G_CALLBACK(item_about_to_show), this);
 			rebuildMenu (subMenu, item);
 		}
-		//TODO: Recreate the "dynamic" popup menus
 		else
 		{
 			//Getting rid of any possible children
@@ -558,140 +604,41 @@ FrameHelper::getLabelFromCommandURL (OUString commandURL)
 	return label;
 }
 
-/*typedef struct {
-	gboolean focused;
-	gchar *  title;
-} WindowInfo;
-
-void
-FrameHelper::populateWindowList(DbusmenuMenuitem *parent)
-{
-	GList *window_list = NULL;
-	Reference < XFramesSupplier > xDesktop(xMSF->createInstance(OUString ( RTL_CONSTASCII_USTRINGPARAM ("com.sun.star.frame.Desktop"))),
-	                                       UNO_QUERY);
-	
-	if (!xDesktop.is())
-		return;
-	
-	//Retreive focus and title info for each toplevel LibreOffice Window
-	Sequence < Reference < XFrame > > xFrames(xDesktop->getFrames()->queryFrames(com::sun::star::frame::FrameSearchFlag::ALL));
-	for (gint32 i = 0; i < xFrames.getLength (); i++)
-	{
-		Reference < XFrame > xFrame (xFrames[i], UNO_QUERY);
-		if (!xFrame.is ())
-			continue;
-
-		if (!xFrame->isTop())
-			continue;
-				
-		Reference < XWindow2 > xWindow (xFrame->getContainerWindow (), UNO_QUERY);
-		if (xWindow.is())
-		{
-			OUString title;
-			WindowInfo *info = (WindowInfo*)g_malloc (sizeof (WindowInfo));
-			
-			Reference< XPropertySet > frameProps (xFrame, UNO_QUERY);
-			frameProps->getPropertyValue (OUString::createFromAscii ("Title")) >>= title;
-			title = title.replace ((sal_Unicode)0x007e, (sal_Unicode)0x005f);
-			
-			//Set the values on the struct and append to the list
-			info->title = g_utf16_to_utf8 (title.getStr(),
-			                               title.getLength(),
-			                               NULL, NULL, NULL);
-			info->focused = xFrame->isActive();
-			
-			window_list = g_list_append (window_list, (gpointer)info);
-		}
-	}
-	
-	//We ensure here that there are enough Menuitems to list all windows
-	//Count the number of items after the last separator:
-	GList *items = dbusmenu_menuitem_get_children (parent);
-	GList *last_sep = g_list_last(items);
-	guint last_sep_index = 0;
-	while (last_sep)
-	{
-		last_sep_index++;
-		last_sep = last_sep->prev;
-		if (!g_strcmp0 (dbusmenu_menuitem_property_get (DBUSMENU_MENUITEM(last_sep->data),
-		                                                "CommandURL"),
-		                "slot:0"));
-		break;
-	}
-	if (!last_sep)
-	{
-		//There was some error looking for the last separator
-		g_list_free_full (window_list, g_free);
-		return;
-	}
-	//Index of the last separator starting from 1
-	last_sep_index = g_list_length (items) - last_sep_index + 1;
-	
-	//Create new menuitems if there are less than we need
-	if ((last_sep_index + g_list_length (window_list)) > g_list_length (items))
-	{
-		// Add enough items to replicate all
-		guint n_new_items = last_sep_index + g_list_length (window_list) - g_list_length (items);
-		for (guint i = 0; i < n_new_items; i++)
-		{
-			DbusmenuMenuitem *item = dbusmenu_menuitem_new ();
-			dbusmenu_menuitem_child_append (parent, item);
-			g_signal_connect(G_OBJECT(item), DBUSMENU_MENUITEM_SIGNAL_ITEM_ACTIVATED, G_CALLBACK(item_activated), this);
-		}
-		items = dbusmenu_menuitem_get_children (parent);
-	}
-
-	last_sep = g_list_last(items);
-	while (last_sep)
-	{
-		last_sep = last_sep->prev;
-		if (!g_strcmp0 (dbusmenu_menuitem_property_get (DBUSMENU_MENUITEM(last_sep->data),
-		                                                "CommandURL"),
-		                "slot:0"));
-		break;
-	}
-	
-
-	//Set the information on all menuitems
-
-	GList *window_info = window_list;
-	GList *window_item = last_sep->next;
-	while (window_item && window_info)
-	{
-		DbusmenuMenuitem *item = DBUSMENU_MENUITEM(window_item->data);
-		WindowInfo *info = (WindowInfo*)window_info->data;
-		
-		//TODO: Find a way to identify the window (setName?)
-		dbusmenu_menuitem_property_set (item, "CommandURL", "focusAnotherWindow");
-		dbusmenu_menuitem_property_set (item, DBUSMENU_MENUITEM_PROP_LABEL, info->title);
-		dbusmenu_menuitem_property_set (item, DBUSMENU_MENUITEM_PROP_TOGGLE_TYPE,  DBUSMENU_MENUITEM_TOGGLE_RADIO);
-		dbusmenu_menuitem_property_set_bool (item, DBUSMENU_MENUITEM_PROP_VISIBLE, TRUE);
-		
-		if (info->focused)
-			dbusmenu_menuitem_property_set_int (item, DBUSMENU_MENUITEM_PROP_TOGGLE_STATE, DBUSMENU_MENUITEM_TOGGLE_STATE_CHECKED);
-		else
-			dbusmenu_menuitem_property_set_int (item, DBUSMENU_MENUITEM_PROP_TOGGLE_STATE, DBUSMENU_MENUITEM_TOGGLE_STATE_CHECKED);
-		
-		window_info = window_info->next;
-		window_item = window_item->next;
-	}
-	
-	g_list_free_full (window_list, g_free);
-}*/
-
 void
 FrameHelper::rebuildMenuFromRoot ()
 {
-	Reference < XFrame > xFrame  = getFrame ();
-	Reference< XPropertySet > frameProps (xFrame, UNO_QUERY);
-	Reference < XLayoutManager > xLayoutManager(frameProps->getPropertyValue(OUString::createFromAscii("LayoutManager")),
-	                                            UNO_QUERY);
-	Reference < XUIElement > menuBar(xLayoutManager->getElement (OUString::createFromAscii("private:resource/menubar/menubar")),
-	                                 UNO_QUERY);
+	Reference < XFrame >         xFrame = getFrame ();
+	Reference < XPropertySet >   frameProps     (xFrame, UNO_QUERY);
+	Reference < XLayoutManager > xLayoutManager (frameProps->getPropertyValue(OUString::createFromAscii("LayoutManager")),
+	                                             UNO_QUERY);
+	Reference < XUIElement >   menuBar     (xLayoutManager->getElement (OUString::createFromAscii("private:resource/menubar/menubar")),
+	                                        UNO_QUERY);
 	Reference < XPropertySet > menuPropSet (menuBar, UNO_QUERY);
 	
-	Reference < XMenu > xMenu(menuPropSet->getPropertyValue(OUString::createFromAscii("XMenuBar")),
-	                          UNO_QUERY);
-
+	Reference < XMenu > xMenu (menuPropSet->getPropertyValue(OUString::createFromAscii("XMenuBar")),
+	                           UNO_QUERY);
 	rebuildMenu (xMenu, root);
 }
+
+gboolean
+FrameHelper::isSpecialSubmenu (OUString command)
+{
+	const gchar * specialSubmenus[11] = { ".uno:CharFontName",
+	                                     ".uno:FontHeight",
+	                                     ".uno:ObjectMenue",
+	                                     ".uno:InsertPageHeader",
+	                                     ".uno:InsertPageFooter",
+	                                     ".uno:ChangeControlType",
+	                                     ".uno:AvailableToolbars",
+	                                     ".uno:ScriptOrganizer",
+	                                     ".uno:RecentFileList",
+	                                     ".uno:AddDirect",
+	                                     ".uno:AutoPilotMenu"};
+
+	for (gint i = 0; i < 11; i++)
+	{
+		if (command.equals (OUString::createFromAscii (specialSubmenus[i])))
+			return TRUE;
+	}
+	return FALSE;
+}
